@@ -24,15 +24,16 @@
  * SUCH DAMAGE.
  */
 
-#ifndef lint
-static const char rcsid[] =
-  "$FreeBSD: src/usr.bin/ncal/ncal.c,v 1.16 2001/06/10 23:26:47 ache Exp $";
-#endif /* not lint */
-
-#include <calendar.h>
+#include "calendar.h"
 #include <ctype.h>
 #include <err.h>
-#include <langinfo.h>
+#ifndef __FreeBSD__
+  #ifndef  __NetBSD__
+    #ifndef __OpenBSD__
+       #include <langinfo.h>
+    #endif
+  #endif
+#endif
 #include <locale.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -40,6 +41,7 @@ static const char rcsid[] =
 #include <sysexits.h>
 #include <time.h>
 #include <unistd.h>
+#include <wchar.h>
 
 /* Width of one month with backward compatibility */
 #define MONTH_WIDTH_B_J 27
@@ -59,13 +61,13 @@ struct monthlines {
 };
 
 struct weekdays {
-	char names[7][4];
+	char *names[7];
 };
 
 /* The switches from Julian to Gregorian in some countries */
 static struct djswitch {
-	char *cc;	/* Country code according to ISO 3166 */
-	char *nm;	/* Name of country */
+	const char *cc;	/* Country code according to ISO 3166 */
+	const char *nm;	/* Name of country */
 	date dt;	/* Last day of Julian calendar */
 } switches[] = {
 	{"AL", "Albania",       {1912, 11, 30}},
@@ -99,7 +101,7 @@ static struct djswitch {
 	{"RO", "Romania",       {1919,  3, 31}},
 	{"RU", "Russia",        {1918,  1, 31}},
 	{"SI", "Slovenia",      {1919,  3,  4}},
-	{"SW", "Sweden",        {1753,  2, 17}},
+	{"SE", "Sweden",        {1753,  2, 17}},
 	{"TR", "Turkey",        {1926, 12, 18}},
 	{"US", "United States", {1752,  9,  2}},
 	{"YU", "Yugoslavia",    {1919,  3,  4}}
@@ -158,14 +160,19 @@ int     flag_weeks;		/* user wants number of week */
 int     nswitch;		/* user defined switch date */
 int	nswitchb;		/* switch date for backward compatibility */
 
-char   *center(char *s, char *t, int w);
+int	weekstart;		/* day the week starts on (Sun [0] - Sat [6]) */
+
+char   *align(char *s, char *t, int w, int l);
+#define center(s, t, w) align((s), (t), (w), 0)
 void	mkmonth(int year, int month, int jd_flag, struct monthlines * monthl);
 void    mkmonthb(int year, int month, int jd_flag, struct monthlines * monthl);
 void    mkweekdays(struct weekdays * wds);
+int     parsemonth(const char *s);
 void    printcc(void);
 void    printeaster(int year, int julian, int orthodox);
 void    printmonth(int year, int month, int jd_flag);
 void    printmonthb(int year, int month, int jd_flag);
+void    printmonthl(int year, int month);
 void    printyear(int year, int jd_flag);
 void    printyearb(int year, int jd_flag);
 int	firstday(int y, int m);
@@ -180,11 +187,12 @@ int
 main(int argc, char *argv[])
 {
 	struct  djswitch *p, *q;	/* to search user defined switch date */
-	date	never = {10000, 1, 1};	/* outside valid range of dates */
+	date	never = {5875706, 1, 1};/* outside valid range of dates */
 	date	ukswitch = {1752, 9, 2};/* switch date for Great Britain */
 	int     ch;			/* holds the option character */
 	int     m = 0;			/* month */
 	int	y = 0;			/* year */
+	int     flag_three = 0;		/* show previous, current, and next month */
 	int     flag_backward = 0;	/* user called cal--backward compat. */
 	int     flag_hole_year = 0;	/* user wants the whole year */
 	int	flag_julian_cal = 0;	/* user wants Julian Calendar */
@@ -193,7 +201,7 @@ main(int argc, char *argv[])
 	int	flag_orthodox = 0;	/* use wants Orthodox easter */
 	int	flag_easter = 0;	/* use wants easter date */
 	char	*cp;			/* character pointer */
-	char    *locale;		/* locale to get country code */
+	const char    *locale;		/* locale to get country code */
 
 	/*
 	 * Use locale to determine the country code,
@@ -236,8 +244,13 @@ main(int argc, char *argv[])
 	if (flag_backward)
 		nswitchb = ndaysj(&ukswitch);
 
-	while ((ch = getopt(argc, argv, "Jejops:wy")) != -1)
+	while ((ch = getopt(argc, argv, "3Jejmops:wy")) != -1)
 		switch (ch) {
+		case '3':
+			if (!flag_backward)
+				usage();
+			flag_three = 1;
+			break;
 		case 'J':
 			if (flag_backward)
 				usage();
@@ -249,6 +262,11 @@ main(int argc, char *argv[])
 				usage();
 			flag_easter = 1;
 			break;
+		case 'm':
+			if (!flag_backward)
+				usage();
+			weekstart = 1;
+			break;
 		case 'j':
 			flag_julian_day = 1;
 			break;
@@ -306,14 +324,16 @@ main(int argc, char *argv[])
 	case 2:
 		if (flag_easter)
 			usage();
-		m = atoi(*argv++);
+		m = parsemonth(*argv++);
 		if (m < 1 || m > 12)
-			errx(EX_USAGE, "month %d not in range 1..12", m);
+			errx(EX_USAGE,
+			    "%s is neither a month number (1..12) nor a name",
+			    argv[-1]);
 		/* FALLTHROUGH */
 	case 1:
 		y = atoi(*argv++);
-		if (y < 1 || y > 9999)
-			errx(EX_USAGE, "year %d not in range 1..9999", y);
+		if (y < 1 || y > 5875706)
+			errx(EX_USAGE, "year %d not in range 1..5875706", y);
 		break;
 	case 0:
 		break;
@@ -328,6 +348,8 @@ main(int argc, char *argv[])
 			printyearb(y, flag_julian_day);
 		else
 			printyear(y, flag_julian_day);
+	else if (flag_three)
+		printmonthl(y, m);
 	else
 		if (flag_backward)
 			printmonthb(y, m, flag_julian_day);
@@ -342,7 +364,7 @@ usage(void)
 {
 
 	fprintf(stderr, "%s\n%s\n%s\n",
-	    "usage: cal [-jy] [[month] year]",
+	    "usage: cal [-3jmy] [[month] year]",
 	    "       ncal [-Jjpwy] [-s country_code] [[month] year]",
 	    "       ncal [-Jeo] [year]");
 	exit(EX_USAGE);
@@ -376,10 +398,14 @@ printeaster(int y, int julian, int orthodox)
 	date    dt;
 	struct tm tm;
 	char    buf[80];
+#ifndef D_MD_ORDER
+	static int d_first = 1; /* XXX */
+#else
 	static int d_first = -1;
 
 	if (d_first < 0)
 		d_first = (*nl_langinfo(D_MD_ORDER) == 'd');
+#endif
 	/* force orthodox easter for years before 1583 */
 	if (y < 1583)
 		orthodox = 1;
@@ -411,7 +437,9 @@ printmonth(int y, int m, int jd_flag)
 	mkweekdays(&wds);
 	printf("    %s %d\n", month.name, y);
 	for (i = 0; i != 7; i++)
-		printf("%.2s%s\n", wds.names[i], month.lines[i]);
+		printf("%s%s\n", wds.names[i], month.lines[i]);
+	for (i = 0; i != 7; i++)
+		free(wds.names[i]);
 	if (flag_weeks)
 		printf("  %s\n", month.weeks);
 }
@@ -434,18 +462,55 @@ printmonthb(int y, int m, int jd_flag)
 	printf("%s\n", center(t, s, mw));
 
 	if (jd_flag)
-		printf(" %s %s %s %s %s %s %.2s\n", wds.names[6], wds.names[0],
-			wds.names[1], wds.names[2], wds.names[3],
+		printf(" %s  %s  %s  %s  %s  %s  %s\n", wds.names[6],
+			wds.names[0], wds.names[1], wds.names[2], wds.names[3],
 			wds.names[4], wds.names[5]);
 	else
-		printf("%s%s%s%s%s%s%.2s\n", wds.names[6], wds.names[0],
+		printf("%s %s %s %s %s %s %s\n", wds.names[6], wds.names[0],
 			wds.names[1], wds.names[2], wds.names[3],
 			wds.names[4], wds.names[5]);
-
+	for (i = 0; i != 7; i++)
+		free(wds.names[i]);
 	for (i = 0; i != 6; i++)
 		printf("%s\n", month.lines[i]+1);
 }
 
+void
+printmonthl(int y, int m)
+{
+	struct monthlines month[3];
+	struct weekdays wds;
+	char s[MAX_WIDTH], t[MAX_WIDTH];
+	int i, n;
+	int mw;
+
+	mkweekdays(&wds);
+	mw = MONTH_WIDTH_B;
+
+	for (n = y * 12 + m - 2, i = 0; i < 3; n++, i++) {
+		mkmonthb(n / 12, n % 12, 0, &month[i]);
+
+		sprintf(s, "%s %d", month[i].name, n / 12);
+		printf("%s  ", center(t, s, mw));
+	}
+	printf("\n");
+
+	for (i = 0; i < 3; i++)
+		printf("%s %s %s %s %s %s %s  ", wds.names[6], wds.names[0],
+			wds.names[1], wds.names[2], wds.names[3],
+			wds.names[4], wds.names[5]);
+	printf("\n");
+	for (i = 0; i < 7; i++)
+		free(wds.names[i]);
+
+	for (i = 0; i < 6; i++) {
+		int j;
+		for (j = 0; j < 3; j++)
+			printf("%s  ", align(t, month[j].lines[i]+1, mw, 1));
+		printf("\n");
+	}
+}
+
 void
 printyear(int y, int jd_flag)
 {
@@ -467,16 +532,16 @@ printyear(int y, int jd_flag)
 
 	for (j = 0; j != 12; j += mpl) {
 		printf("    %-*s%-*s",
-		    mw, year[j].name,
-		    mw, year[j + 1].name);
+		    mw, align(s, year[j].name, mw, 1),
+		    mw, align(t, year[j + 1].name, mw, 1));
 		if (mpl == 3)
 			printf("%s\n", year[j + 2].name);
 		else
 			printf("%-*s%s\n",
-		    	    mw, year[j + 2].name,
+		    	    mw, align(t, year[j + 2].name, mw, 1),
 		    	    year[j + 3].name);
 		for (i = 0; i != 7; i++) {
-			printf("%.2s%-*s%-*s",
+			printf("%s%-*s%-*s",
 			    wds.names[i],
 			    mw, year[j].lines[i],
 			    mw, year[j + 1].lines[i]);
@@ -501,6 +566,8 @@ printyear(int y, int jd_flag)
 				    year[j + 3].weeks);
 		}
 	}
+	for (i = 0; i != 7; i++)
+		free(wds.names[i]);
 }
 
 void
@@ -532,8 +599,8 @@ printyearb(int y, int jd_flag)
 			    center(t, year[j + 2].name, mw));
 
 		if (mpl == 2)
-			printf(" %s %s %s %s %s %s %s "
-			       " %s %s %s %s %s %s %.2s\n",
+			printf(" %s  %s  %s  %s  %s  %s  %s  "
+			       " %s  %s  %s  %s  %s  %s  %s\n",
 				wds.names[6], wds.names[0], wds.names[1],
 				wds.names[2], wds.names[3], wds.names[4],
 				wds.names[5],
@@ -541,9 +608,9 @@ printyearb(int y, int jd_flag)
 				wds.names[2], wds.names[3], wds.names[4],
 				wds.names[5]);
 		else
-			printf("%s%s%s%s%s%s%s "
-				"%s%s%s%s%s%s%s "
-				"%s%s%s%s%s%s%.2s\n",
+			printf("%s %s %s %s %s %s %s  "
+				"%s %s %s %s %s %s %s  "
+				"%s %s %s %s %s %s %s\n",
 				wds.names[6], wds.names[0], wds.names[1],
 				wds.names[2], wds.names[3], wds.names[4],
 				wds.names[5],
@@ -566,6 +633,8 @@ printyearb(int y, int jd_flag)
 
 		}
 	}
+	for (i = 0; i != 7; i++)
+		free(wds.names[i]);
 }
 
 void
@@ -587,7 +656,7 @@ mkmonth(int y, int m, int jd_flag, struct monthlines *mlines)
 	/* Set name of month. */
 	memset(&tm, 0, sizeof(tm));
 	tm.tm_mon = m;
-	strftime(mlines->name, sizeof(mlines->name), "%OB", &tm);
+	strftime(mlines->name, sizeof(mlines->name), "%B", &tm);
 	mlines->name[0] = toupper((unsigned char)mlines->name[0]);
 
 	/*
@@ -679,7 +748,7 @@ mkmonthb(int y, int m, int jd_flag, struct monthlines *mlines)
 	/* Set name of month centered */
 	memset(&tm, 0, sizeof(tm));
 	tm.tm_mon = m;
-	strftime(mlines->name, sizeof(mlines->name), "%OB", &tm);
+	strftime(mlines->name, sizeof(mlines->name), "%B", &tm);
 	mlines->name[0] = toupper((unsigned char)mlines->name[0]);
 
 	/*
@@ -710,10 +779,10 @@ mkmonthb(int y, int m, int jd_flag, struct monthlines *mlines)
 	}
 
 	/*
-	 * Set firsts to the day number of sunday of the first week of
-	 * this month. (This might be in the last month)
+	 * Set firsts to the weekstart day of the first week of this
+	 * month as a day number. (This might be in the last month.)
 	 */
-	firsts = first - (weekday(first)+1) % 7;
+	firsts = first - (weekday(first)+(8-weekstart)) % 7;
 
 	/*
 	 * Fill the lines with day of month or day of year (Julian day)
@@ -746,17 +815,28 @@ mkweekdays(struct weekdays *wds)
 	int i, len;
 	struct tm tm;
 	char buf[20];
+	int j, col;
+	wchar_t *wcs;
 
 	memset(&tm, 0, sizeof(tm));
 
 	for (i = 0; i != 7; i++) {
-		tm.tm_wday = (i+1) % 7;
+		tm.tm_wday = (i+1+weekstart) % 7;
 		strftime(buf, sizeof(buf), "%a", &tm);
-		len = strlen(buf);
-		if (len > 2)
-			len = 2;
-		strcpy(wds->names[i], "   ");
-		strncpy(wds->names[i] + 2 - len, buf, len);
+		len = mbstowcs(NULL, buf, 0);
+		wcs = calloc(len + 1, sizeof(wchar_t));
+		mbstowcs(wcs, buf, len);
+		for (j = 0; j < len; j++)
+			if (wcswidth(wcs, j + 1) >= 2)
+				break;
+		wcs[j + 1] = L'\0';
+		col = wcswidth(wcs, j + 1);
+		len = wcstombs(NULL, wcs, 0);
+		wds->names[i] = calloc(len + 2 - col + 1, sizeof(char));
+		wcstombs(wds->names[i], wcs, len);
+		free(wcs);
+		if (col < 2)
+			strcat(wds->names[i], " ");
 	}
 }
 
@@ -838,13 +918,42 @@ sdateb(int nd, struct date *d)
 		return (jdate(nd, d));
 }
 
-/* Center string t in string s of length w by putting enough leading blanks */
+/*
+ * Center (or left-align, if l != 0) string t in string s of length w by
+ * putting enough leading and trailing blanks
+ */
 char *
-center(char *s, char *t, int w)
+align(char *s, char *t, int w, int l)
 {
+	wchar_t buf[MAX_WIDTH + 1];
 	char blanks[80];
+	int n, r, a, b;
+
+	if ((n = mbstowcs(buf, t, MAX_WIDTH)) == -1 ||
+	    (r = wcswidth(buf, n)) == -1)
+		r = strlen(t);
+
+	a = l ? 0 : (w - r) / 2;
+	b = w - a - r;
 
 	memset(blanks, ' ', sizeof(blanks));
-	sprintf(s, "%.*s%s", (int)(w - strlen(t)) / 2, blanks, t);
+	sprintf(s, "%.*s%s%.*s", a, blanks, t, b, blanks);
 	return (s);
 }
+
+int
+parsemonth(const char *s)
+{
+	int v;
+	char *cp;
+	struct tm tm;
+
+	v = (int)strtol(s, &cp, 10);
+	if (cp != s)
+		return (v);
+	if (strptime(s, "%B", &tm) != NULL)
+		return (tm.tm_mon + 1);
+	if (strptime(s, "%b", &tm) != NULL)
+		return (tm.tm_mon + 1);
+	return (0);
+}
